package flow.components

import chisel3._
import chisel3.experimental.noPrefix
import chisel3.util.HasBlackBoxPath
import chisel3.util.HasBlackBoxResource
import chisel3.util.log2Ceil
import chisel3.util.BitPat
import flow.components
import flow.Params
import shapeless.::
import shapeless.HNil

import scala.collection.SeqMap
import chisel3.util.experimental.decode.{decoder, TruthTable}
import flow.components.util._
import flow.components.RV32Inst._

class RamControlInterface(addrWidth: Int) extends Bundle {
  val valid = Input(Bool())
  val writeMask = Input(UInt((addrWidth / 8).W))
  val writeEnable = Input(Bool())
  def ctrlBindPorts = {
    valid :: writeMask :: writeEnable :: HNil
  }

}

/* FIXME: Extends here might not be the best solution.
 *        We need a way to merge two bundles together
 */
class RamInterface[T <: Data](tpe: T, addrWidth: Int)
    extends RamControlInterface(addrWidth) {
  val clock = Input(Clock())
  val reset = Input(Reset())
  val writeAddr = Input(UInt(addrWidth.W))
  val writeData = Input(tpe)
  val readAddr = Input(UInt(addrWidth.W))
  val readData = Output(tpe)
  val pc = Input(UInt(addrWidth.W))
  val inst = Output(tpe)
}

class RamDpi extends BlackBox with HasBlackBoxResource {
  val io = IO((new RamInterface(UInt(32.W), 32)))
  addResource("/RamDpi.v")
}

class DpiRamControlInterface(implicit p: Params) extends Bundle {
  val valid = Input(Bool())
  val writeMask = Input(UInt((p.XLEN.get / 8).W))
  val writeEnable = Input(Bool())
}

class DpiRamInterface(implicit p: Params) extends DpiRamControlInterface {
  val clock = Input(Clock())
  val reset = Input(Reset())

  // Data
  val writeAddr = Input(UInt(p.XLEN))
  val writeData = Input(UInt(p.XLEN))
  val readAddr = Input(UInt(p.XLEN))
  val readData = Output(UInt(p.XLEN))
  val pc = Input(UInt(p.XLEN))
  val inst = Output(UInt(p.XLEN))
}

class RamController(implicit p: Params) extends Module {
  val in = new Bundle {
    val inst = IO(Flipped(UInt(p.XLEN)))
  }
  val out = IO(Flipped(new DpiRamControlInterface))
  private val validMapping = TruthTable(
    Array(sb, sh, sw, lb, lbu, lh, lhu, lw).map(_ -> BitPat("b1")),
    BitPat("b0")
  )

  private val writeMaskMapping = TruthTable(
    Array(
      sb -> BitPat("b0001"),
      sh -> BitPat("b0011"),
      sw -> BitPat("b1111")
    ),
    BitPat.dontCare(out.writeMask.getWidth)
  )

  private val writeEnableMapping = TruthTable(
    Array(
      sb -> BitPat("b1"),
      sh -> BitPat("b1"),
      sw -> BitPat("b1")
    ),
    BitPat("b0")
  )

  out.valid := decoder(in.inst, validMapping)
  out.writeEnable := decoder(in.inst, writeMaskMapping)
  out.writeMask := decoder(in.inst, writeMaskMapping)
}
